/****************************************************************************
*   serial.c -- Serial device handling routines                             *
*   Copyright (C) 1998-1999  Peter Mack, DL3FCJ                             *
*                                                                           *
*   This program is free software; you can redistribute it and/or modify    *
*   it under the terms of the GNU General Public License as published by    *
*   the Free Software Foundation; either version 2 of the License, or       *
*   (at your option) any later version.                                     *
*                                                                           *
*   This program is distributed in the hope that it will be useful,         *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
*   GNU General Public License for more details.                            *
*                                                                           *
*   You should have received a copy of the GNU General Public License       *
*   along with this program; if not, write to the                           *
*   Free Software Foundation, Inc.                                          *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.                *
*                                                                           *
****************************************************************************/

#include "serial.h"


/*
 * global variables
 */
static PSERDATA		pserialdata = NULL;
static int		numsers = 0;


/*
 * Open RS232 interface
 */
int rs232_open(char *device, int speed)
{
	int		handle;
	struct termios	new_termios;
	PSERDATA	pserdata;
	
	if (lock_device(device) < 0) {
		printf("Device locked.\n");
		return -1;
	}

	if ((handle = open(device, O_RDWR)) == -1) {
		fprintf(stderr, "Error: can't open serial port\n");
		unlock_device(device);
		return -1;
	}
	
	if (!numsers) {
		/* allocate the first element */
		pserialdata = calloc (1, sizeof(SERDATA));
		pserdata = pserialdata;
	} else {
		/* search the last element in the list */
		pserdata = pserialdata;
		while (pserdata->Next != NULL) {
			pserdata = pserdata->Next;
		}
		/* allocate memory for the new element */
		pserdata->Next = calloc(1, sizeof(SERDATA));
		pserdata->Next->Prev = pserdata;
		pserdata = pserdata->Next;
	}
	
	pserdata->device = device;
	pserdata->devicehandle = handle;
	
	numsers++;
	
	/* save current port status */
	tcgetattr(handle, &pserdata->old_termios);

	/*
	 * set the baudrate
	 */
	switch (speed) {
		case 300:
			speed = B300;
			break;
		case 600:
			speed = B600;
			break;
		case 1200:
			speed = B1200;
			break;
		case 2400:
			speed = B2400;
			break;
		case 4800:
			speed = B4800;
			break;
		case 9600:
			speed = B9600;
			break;
		case 19200:
			speed = B19200;
			break;
		case 38400:
			speed = B38400;
			break;
		case 57600:
			speed = B57600;
			break;
		case 115200:
			speed = B115200;
			break;
		default:
			speed = B38400;
	}

	new_termios = pserdata->old_termios;
	
	new_termios.c_cc[VTIME] = 0;
	new_termios.c_cc[VMIN] = 1;
	new_termios.c_iflag = 0;
	new_termios.c_iflag |= IGNBRK;
	new_termios.c_oflag = 0;
	new_termios.c_lflag = 0;
	new_termios.c_cflag |= (CS8 | CREAD | CLOCAL);
	new_termios.c_cflag &= ~(CSTOPB | PARENB | PARODD | HUPCL);
	
	cfsetispeed(&new_termios, speed);
	cfsetospeed(&new_termios, speed);
	
	tcsetattr(handle, TCSADRAIN, &new_termios);
	
	return handle;
}


/*
 * Close RS232 interface
 */
void rs232_close(int handle)
{
	PSERDATA	pserdata;

	/*
	 * Do only if there are active devices
	 */
	if (numsers) {
		if ((pserdata = find_device(handle)) != NULL) {
			tcsetattr(handle, TCSADRAIN, &pserdata->old_termios);
			close(handle);
 			unlock_device(pserdata->device);
			
			if (pserdata->Prev != NULL)
				if (pserdata->Next != NULL) {
					pserdata->Next->Prev = pserdata->Prev;
					pserdata->Prev->Next = pserdata->Next;
				} else {
					pserdata->Prev->Next = NULL;
				}
			else
				if (pserdata->Next != NULL) {
					pserdata->Next->Prev = NULL;
					pserialdata = pserdata->Next;
				} else {
					pserialdata = NULL;
				}

			free(pserdata);

			numsers--;
		}
	}
}


/*
 * Read from RS232 interface
 */
int rs232_read(int handle, void *bp, int maxlen)
{
	fd_set		set;
	struct timeval	timeout;
	int		Max;
	int		endloop;
	int		res;

	
	Max = 0;
	endloop = 0;
	do {
		timeout.tv_sec = 0;
		timeout.tv_usec = 500000; /* 0.5 seconds */
		
		FD_ZERO(&set);
		FD_SET(handle, &set);
		res = select(FD_SETSIZE, &set, NULL, NULL, &timeout);
		if (0 == res) {
			fprintf(stderr, "ERROR: timed out!\n");
			endloop = 1;
		} else if (-1 == res) {
			perror("rs232_read select");
			exit(10);
		} else {
			Max += read(handle, bp, maxlen);
	
		}
	} while ((Max < maxlen) && !endloop);
	
	return res;
}


/*
 * Read from RS232 interface
 */
int rs232_readflush(int handle)
{
	fd_set		set;
	struct timeval	timeout;
	int		Max;
	int		endloop;
	int		res;
	char		tmp;

	
	Max = 0;
	endloop = 0;
	do {
		timeout.tv_sec = 0;
		timeout.tv_usec = 100000; /* 0.1 seconds */
		
		FD_ZERO(&set);
		FD_SET(handle, &set);
		res = select(FD_SETSIZE, &set, NULL, NULL, &timeout);
		if (0 == res) {
//			fprintf(stderr, "ERROR: timed out!\n");
			endloop = 1;
		} else if (-1 == res) {
			perror("rs232_read select");
			exit(10);
		} else {
			Max += read(handle, &tmp, 1);
	
		}
	} while (!endloop);
	
	return res;
}


/*
 * Flush RS232
 */
int rs232_Flush(int handle)
{
	int	blkmode;
	char	buf;
	int	i;
	
	blkmode = fcntl(handle, F_GETFL, NULL);
	fcntl(handle, F_SETFL, (O_NONBLOCK | blkmode));
	i = 0;
	while (read(handle, &buf, 1) > 0) {
		usleep(20000);
		i++;
	}
	printf("%d characters read.\n", i);

	fcntl(handle, F_SETFL, blkmode);

	return(0);
}


/*
 * Get RS232 status
 */
int rs232_Status(int handle)
{
	int	status;

	ioctl(handle, TIOCMGET, &status);

	return status;
}


/*
 * Set or reset DTR
 */
void rs232_DTR(int handle, char l)
{
	int	bit = TIOCM_DTR;
	
	if (l) {
		ioctl(handle, TIOCMBIS, &bit); /* set DTR   */
	} else {
		ioctl(handle, TIOCMBIC, &bit); /* clear DTR */
	}
}


/*
 * Set or reset RTS
 */
void rs232_RTS(int handle, char l)
{
	int	bit = TIOCM_RTS;
	
	if (l) {
		ioctl(handle, TIOCMBIS, &bit); /* set RTS   */
	} else {
		ioctl(handle, TIOCMBIC, &bit); /* clear RTS */
	}
}


int rs232_GetCTS(int handle)
{
	int	status;
	
	ioctl(handle, TIOCMGET, &status);
	
	return (status & TIOCM_CTS);
}


int rs232_GetDSR(int handle)
{
	int	status;
	
	ioctl(handle, TIOCMGET, &status);
	
	return (status & TIOCM_DSR);
}


int rs232_GetDCD(int handle)
{
	int	status;
	
	ioctl(handle, TIOCMGET, &status);
	
	return (status & TIOCM_CAR);
}


int rs232_GetRI(int handle)
{
	int	status;
	
	ioctl(handle, TIOCMGET, &status);
	
	return (status & TIOCM_RNG);
}


/*
 * Find device data in the list
 */
PSERDATA find_device(int handle)
{
	PSERDATA	pserdata;
	int		i;

	for (i = 0, pserdata = pserialdata; i < numsers; i++, pserdata = pserdata->Next)
		if (pserdata->devicehandle == handle)
			return pserdata;
	return NULL;
}

