/****************************************************************************
*   update.c -- PTC-II update handling                                      *
*   Copyright (C) 1998-1999  Peter Mack, DL3FCJ                             *
*                                                                           *
*   This program is free software; you can redistribute it and/or modify    *
*   it under the terms of the GNU General Public License as published by    *
*   the Free Software Foundation; either version 2 of the License, or       *
*   (at your option) any later version.                                     *
*                                                                           *
*   This program is distributed in the hope that it will be useful,         *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
*   GNU General Public License for more details.                            *
*                                                                           *
*   You should have received a copy of the GNU General Public License       *
*   along with this program; if not, write to the                           *
*   Free Software Foundation, Inc.                                          *
*   59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.                *
*                                                                           *
****************************************************************************/

#include "update.h"
#include "serial.h"


int Update(int ser, char *UpdateFileName)
{
	char		Buffer[MAX_SECT_SIZE];

	char		ch;
	char		ManCode;
	char		DevId;
	FLASH		Flash;
	unsigned long	FlashFree;
	unsigned short	Sectors;
	unsigned long	SectorsWritten = 0;
	unsigned long	BytesRead;

	int		hFile;
	unsigned long	FileLength;

	int		res;
	
	FDTIME		FileStamp;
	FDTIME		FlashStamp;
	

	hFile = open(UpdateFileName, O_RDONLY);

	if(-1 == hFile) {
		fprintf(stderr, "ERROR: opening file: %s\n", UpdateFileName);
		return -1;
	}

	lseek(hFile, 12, SEEK_SET);
	read(hFile, &FileStamp, 4);
	
	FileLength = lseek(hFile, 0, SEEK_END);

	res = write(ser, "Q\r", 2);
	usleep(200);

	write(ser, "UPDATE\r", 7);

	rs232_readflush(ser); /* read and ignore the UPDATE message */

	write(ser, "\006", 1);	/* send ACK */

	rs232_read(ser, &ManCode, 1);
	rs232_read(ser, &DevId, 1);
	rs232_read(ser, &FlashStamp, 4);

	if(GetFlash(ManCode, DevId, &Flash) != 0) {
		fprintf(stderr, "ERROR: receiving Flash information!\n");
		write(ser, "\033", 1);	/* send ESC */
		close(hFile);
		return -1;
	}

	printf("Flash: %s - %s\n\n", Flash.Manufacturer, Flash.Type);

	FlashFree = Flash.ulFlashSize - 16384 - Flash.usSectSize;

	if(Flash.usSectSize == 0) {
		fprintf(stderr, "ERROR: wrong sector size!\n");
		write(ser, "\033", 1);	/* send ESC */
		close(hFile);
		return -1;
	}

	Sectors = FileLength / Flash.usSectSize;

	if(FileLength % Flash.usSectSize) {
		Sectors++;
	}

	if(FileLength > FlashFree) {
		fprintf(stderr, "ERROR: File too large!\n       File should not be longer than %ld bytes.\n", FlashFree);
		write(ser, "\033", 1);	/* send ESC */
		close(hFile);
		return -1;
	}
	
	if(FlashStamp.day == 0 || FlashStamp.month == 0 || (FlashStamp.day == 0x1f && FlashStamp.month == 0xf && FlashStamp.year == 0x7f)) {
		fprintf(stderr, "WARNING: Invalid Flash time stamp.\n         Possibly no firmware installed.\n\n");
	}

	if ( convtime(FileStamp) <= convtime(FlashStamp)) {
		printf("The current firmware has the same or a newer time stamp!\n");
		printf("Press <P> to proceed or any other key to quit.\n\n");
		
		do {
			res = fgetc(stdin);
		} while (res == '\n' || res == '\r');
		
		if('p' != (char)res && 'P' != (char)res) {
			write(ser, "\033", 1);	/* send ESC */
			close(hFile);
			return -2;
		}
	}


	write (ser, "\006", 1);	/* send ACK */
	
	ch = (char) (Sectors >> 8);
	write (ser, &ch, 1);
	ch = (char) Sectors;
	write (ser, &ch, 1);

	rs232_read(ser, &ch, 1);

	if (ch != ACK) {
		fprintf(stderr, "\a\aERROR: Handshake failed!\n");
		write(ser, "\033", 1);	/* send ESC */
		close(hFile);
		return -1;
	}

	printf("Writing %ld byte in %d sectors.\n\n", FileLength, Sectors);
	
	lseek(hFile, 0, SEEK_SET);

	do {
		BytesRead = read(hFile, &Buffer, Flash.usSectSize);
		if(BytesRead < Flash.usSectSize) {
			memset(Buffer+BytesRead, 0, Flash.usSectSize-BytesRead);
		}
		write(ser, &Buffer, Flash.usSectSize);
		SectorsWritten++;
		
		rs232_read(ser, &ch, 1);
		
		if (ch != ACK) {
			fprintf(stderr, "\a\aERROR: Handshake failed!\n");
			write(ser, "\033", 1);	/* send ESC */
			close(hFile);
			return -1;
		}
		
		sprintf(Buffer, "Written: %3ld%%\r", SectorsWritten * 100 / Sectors);
		write(STDOUT_FILENO, Buffer, strlen(Buffer));
		
	} while (BytesRead == Flash.usSectSize);

	write(ser, "\r", 1);

	printf("\n\n\aUpdate complete.\n");
	
	close(hFile);

	return 0;
}


int GetFlash(char ManCode, char DevId, FLASH *Flash)
{
	int Result;

	switch(ManCode) {
		/* AMD Flash wird zwar erkannt aber fr Programmierung nicht untersttzt */
		case 0x01:
			Flash->Manufacturer = "AMD";
			switch(DevId) {
				case 0xa4:
					Flash->Type = "Am29F040";
					Flash->usSectSize = 65535;
					Flash->ulFlashSize = 524288;
					Result = 0;
					break;
				default:
					Flash->Type = "UNKNOWN";
					Flash->usSectSize = 0;
					Flash->ulFlashSize = 0;
					Result = -1;
					break;
				}
				break;

		case 0x1f:
			Flash->Manufacturer = "Atmel";
			switch(DevId) {
				case 0x5d:
					Flash->Type = "AT29C512";
					Flash->usSectSize = 128;
					Flash->ulFlashSize = 65535;
					Result = 0;
					break;
				case 0xd5:
					Flash->Type = "AT29C010";
					Flash->usSectSize = 128;
					Flash->ulFlashSize = 131072;
					Result = 0;
					break;
				case 0xda:
					Flash->Type = "AT29C020";
					Flash->usSectSize = 256;
					Flash->ulFlashSize = 262144;
					Result = 0;
					break;
				case 0x5b:
					Flash->Type = "AT29C040";
					Flash->usSectSize = 512;
					Flash->ulFlashSize = 524288;
					Result = 0;
					break;
				case 0xa4:
					Flash->Type = "AT29C040A";
					Flash->usSectSize = 256;
					Flash->ulFlashSize = 524288;
					Result = 0;
					break;
				default:
					Flash->Type = "UNKNOWN";
					Flash->usSectSize = 0;
					Flash->ulFlashSize = 0;
					Result = -1;
					break;
				}
				break;

		default:
			Flash->Manufacturer = "INVALID";
			Flash->Type = "INVALID";
			Flash->usSectSize = 0;
			Flash->ulFlashSize = 0;
			Result = -1;
			break;
		}

	return Result;
}


time_t convtime(FDTIME PTC_Time)
{
	struct tm	btime;
	
	btime.tm_sec = PTC_Time.twosecs * 2;
	btime.tm_min = PTC_Time.minutes;
	btime.tm_hour = PTC_Time.hours;
	btime.tm_mday = PTC_Time.day;
	btime.tm_mon = PTC_Time.month;
	btime.tm_year = PTC_Time.year + 80;
	btime.tm_wday = 0;
	btime.tm_yday = 0;
	btime.tm_isdst = 0;

	return mktime(&btime);
}
